#!/usr/bin/env python
import sys
import os
import urllib
import gzip
import errno
import re
from glob import glob, iglob

from prosci.common import NotFoundError, ParsingError
from prosci.util.protein import Protein, Pdb
from prosci.util.getpdb import getpdb
from prosci.loops.freaddb import FreadDB


# Taken from:
# http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
def mkdir_p(path):
  try:
    os.makedirs(path)
  except OSError as exc:
    if exc.errno == errno.EEXIST:
      pass
    else:
      raise



#########################################################################



def set_of_potential_membrane_proteins():
  pdb_codes = set([])
  
  def addall(lst):
    for x in lst:
      pdb_codes.add(x.upper())
  
  
  # White lab's Database of Membrane Protein Structures
  f = urllib.urlopen("http://blanco.biomol.uci.edu/mpstruc/listAll/mpstrucTblXml")
  txt = f.read()
  f.close()
  addall(re.findall("<pdbCode>(.*)</pdbCode>", txt))


  # OPM
  f = urllib.urlopen("http://opm.phar.umich.edu/download.php?see=pdbids")
  txt = f.read()
  f.close()
  for line in txt.splitlines():
    if "PDB entries" in line:
      fields = line.split("<br />")
      for x in fields:
        if len(x) == 4:
          pdb_codes.add(x.upper())


  # PDB_TM
  f = urllib.urlopen("http://pdbtm.enzim.hu/data/pdbtm_all.list")
  txt = f.read()
  f.close()
  for line in txt.splitlines():
    x = line[:4]
    if x:
      pdb_codes.add(x.upper())


  f = urllib.urlopen("http://sbcb.bioch.ox.ac.uk/cgdb/dbstart.php")
  txt = f.read()
  f.close()
  addall(re.findall("pdb=(\w+)", txt))
  
  return pdb_codes


def set_of_alpha_membrane_chains():
  pdb_codes = set([])
  
  # PDB_TM
  f = urllib.urlopen("http://pdbtm.enzim.hu/data/pdbtm_alpha.list")
  txt = f.read()
  f.close()
  for line in txt.splitlines():
    x = line[:4].upper()+line[-1:]
    if x:
      pdb_codes.add(x)
  
  return pdb_codes


def set_of_beta_membrane_chains():
  pdb_codes = set([])
  
  # PDB_TM
  f = urllib.urlopen("http://pdbtm.enzim.hu/data/pdbtm_beta.list")
  txt = f.read()
  f.close()
  for line in txt.splitlines():
    x = line[:4].upper()+line[-1:]
    if x:
      pdb_codes.add(x)
  
  return pdb_codes



def write_1dlist_file(fname, data):
  f = open(fname, "wb")
  for item in data:
    f.write(item)
    f.write("\n")
  f.close()

def read_1dlist_file(fname):
  f = open(fname, "rb")
  data = []
  for line in f:
    data.append(line.rstrip())
  f.close()
  return data

def write_2dlist_file(fname, data):
  f = open(fname, "wb")
  for item in data:
    f.write(" ".join(item))
    f.write("\n")
  f.close()


def get_putative_protein_list(fpath, func):
  if not os.path.exists(fpath):
    data = func()
    write_1dlist_file(fpath, sorted(data))
  else:
    data = set(read_1dlist_file(fpath))
  return data



#########################################################################



def getchains(pdbcode, chaincodes=None, dbdir=None):
    "Get PDB structures as Pdb objects"
    
    pdbcode = pdbcode.lower()
    a = None
    
    if dbdir is not None:
      try:
        a = getpdb(pdbcode, dbdir)
      except NotFoundError:
        pass
    
    if a is None:
      fname, info = urllib.urlretrieve("http://www.pdb.org/pdb/files/%s.pdb.gz" % (pdbcode.upper()))
      try:
        gf = gzip.open(fname)
        try:
          a = Pdb(pdbcode, gf)
        except IOError:
          raise NotFoundError("Could not retrieve pdb '%s' from www.pdb.org"%(pdbcode))
        except ParsingError:
          raise NotFoundError("Could not retrieve pdb '%s' from www.pdb.org"%(pdbcode))
        finally:
          gf.close()
      finally:
        if os.path.exists(fname):
          os.remove(fname)
    
    #chains = a.get_chain_codes()
    if not chaincodes:
      return a
    return a.get_chain(chaincodes, include_duplicates=False)



#########################################################################



minLength = 40
maxResolution = 3.0
maxR = 0.3
maxRfree = 0.35
validMethod = "XRAY"


if __name__ == "__main__":
  
  pdbdir, = sys.argv[1:]
  
  
  putative_membrane_protein_pdbs = get_putative_protein_list("putative_membrane_protein_pdbs.list", set_of_potential_membrane_proteins)
  
  putative_membrane_alpha_chains = get_putative_protein_list("putative_membrane_alpha_chains.list", set_of_alpha_membrane_chains)
  
  putative_membrane_beta_chains = get_putative_protein_list("putative_membrane_beta_chains.list", set_of_beta_membrane_chains)
  
  
  if not os.path.exists("pdbaanr.gz"):
    urllib.urlretrieve("http://dunbrack.fccc.edu/Guoli/culledpdb/pdbaanr.gz", "pdbaanr.gz")

  sol_complex_chains = []
  mem_complex_chains = []
  mem_alpha_chains = []
  mem_beta_chains = []
  mem_other_chains = []
  
  f = gzip.open("pdbaanr.gz")
  try:
    for line in f:
      if not line.startswith(">"):
        continue
      
      line = line[1:].rstrip()
      
      try:
        fields = line.split("||", 1)
        if len(fields) == 1:
          info = fields[0]
          aliases = ""
        else:
          info, aliases = fields
        
        code, length, method, resolution, R, Rfree, unknown_boolean, name = info.split(None, 7)
        aliases = aliases.split()
        aliases.insert(0, code)
        
        if resolution == "NA":
          resolution = sys.maxint
        if R == "NA":
          R = 1.0
        if Rfree == "NA":
          Rfree = 1.0
        
        if (method == validMethod
          and int(length) >= minLength
          and float(resolution) <= maxResolution
          and float(R) <= maxR
          and float(Rfree) <= maxRfree):
            pdb = code[:4].upper()
            if pdb in putative_membrane_protein_pdbs:
              mem_complex_chains.append(aliases)
              isalpha = False
              for chain in aliases:
                if chain in putative_membrane_alpha_chains:
                  mem_alpha_chains.append(aliases)
                  isalpha = True
                  break
              if not isalpha:
                isbeta = False
                for chain in aliases:
                  if chain in putative_membrane_beta_chains:
                    mem_beta_chains.append(aliases)
                    isbeta = True
                    break
                if not isbeta:
                  mem_other_chains.append(aliases)
            else:
             sol_complex_chains.append(aliases)
      except:
        print line
        raise
  finally:
    f.close()
  
  sol_complex_chains.sort()
  write_2dlist_file("sol_complex_chains.list", sol_complex_chains)

  mem_complex_chains.sort()
  write_2dlist_file("mem_complex_chains.list", mem_complex_chains)

  mem_alpha_chains.sort()
  write_2dlist_file("mem_alpha_chains.list", mem_alpha_chains)

  mem_beta_chains.sort()
  write_2dlist_file("mem_beta_chains.list", mem_beta_chains)
  
  mem_other_chains.sort()
  write_2dlist_file("mem_other_chains.list", mem_other_chains)
  
  
  databases = [("sol", sol_complex_chains),
               ("memalpha", mem_alpha_chains),
               ("membeta", mem_beta_chains),
               ("memother", mem_other_chains)]
  
  min_loop_length = 3
  max_loop_length = 30
  build_structure_collection = False
  build_indeces = True
  
  for db, contents in databases:
    dbdir = "db_"+db
    
    freaddb = FreadDB(dbdir)
    freaddb.verbose = True
    freaddb.terminal_cutoff = 5
    freaddb.save_db_options()
    
    if build_structure_collection:
      current_pdb = ""
      current_struc = None
      selected_chains = []
      for ichain, chainlist in enumerate(contents):
        code = chainlist[0]
        pdb = code[:4].lower()
        chaincode = code[4:5]
        if pdb != current_pdb:
          if current_pdb:
            print "%d/%d chains: %s" % (ichain+1, len(contents), pdb)
            if current_struc is not None:
              freaddb.add_structure(current_struc, chain_codes=selected_chains, skip_existing=True)
          current_pdb = pdb
          current_struc = None
          if not glob(os.path.join(freaddb.strucdir, pdb[1:3], pdb+"*.tem")):
            try:
              current_struc = Protein(getchains(pdb, dbdir = dbdir))
            except NotFoundError, e:
              print "       ", e
          selected_chains = [chaincode]
        else:
          selected_chains.append(chaincode)
    
    if build_indeces:
      
      proteins_missing_annotation = set([])
      for fname in iglob(os.path.join(freaddb.strucdir, "*", "*.atm")):
        if not os.path.exists(fname[-len(".atm.gz"):]+".tem"):
          proteins_missing_annotation.add(os.path.basename(fname)[:4].lower())
      
      if proteins_missing_annotation:
        filtered_contents = []
        for ichain, chainlist in enumerate(contents):
          code = chainlist[0]
          pdb = code[:4].lower()
          if pdb in proteins_missing_annotation:
            filtered_contents.append(chainlist)
            proteins_missing_annotation.remove(pdb)
            if not proteins_missing_annotation:
              break
        contents = filtered_contents
      else:
        contents = []
      
      # Re-add structures that were missing annotations
      # Copy/pasted from above, except for skip_existing=False
      current_pdb = ""
      current_struc = None
      selected_chains = []
      for ichain, chainlist in enumerate(contents):
        code = chainlist[0]
        pdb = code[:4].lower()
        chaincode = code[4:5]
        if pdb != current_pdb:
          if current_pdb:
            print "%d/%d chains: %s" % (ichain+1, len(contents), pdb)
            if current_struc is not None:
              freaddb.add_structure(current_struc, chain_codes=selected_chains, skip_existing=False)
          current_pdb = pdb
          current_struc = None
          if not glob(os.path.join(freaddb.strucdir, pdb[1:3], pdb+"*.tem")):
            try:
              current_struc = Protein(getchains(pdb, dbdir = dbdir))
            except NotFoundError, e:
              print "       ", e
          selected_chains = [chaincode]
        else:
          selected_chains.append(chaincode)
      
      for loop_length in range(min_loop_length, max_loop_length+1):
        print "Build database for loop length %d" % (loop_length)
        freaddb.build_index(loop_length)
