#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# iMembrane : Annotation database construction module
#
# Author: Sebastian Kelm
# Created: 28/01/2008
#
# Revisions:
# 29/01/2008    Sebastian Kelm    Initial version
# 16/04/2008    Sebastian Kelm    Conversion to object-oriented code
# 30/04/2008    Sebastian Kelm    Merger with ALI file handling
# 01/02/2010    Sebastian Kelm    Split from ALI. Now includes rewritten structre handling (no longer needs PyMOL).
# 06/05/2010    Sebastian Kelm    Split from annotation projection code. Now only builds an initial backend database using CGDB data.
#





import sys, re
from array import array
from glob import glob

from prosci.common import *
from prosci.util.pdb import Pdb, validate_pdb_code
from prosci.util.ali import Ali
from prosci.util.gaps import gappify, deGappify, isGap


DEFAULT_TIMELIMIT    = 0.1   # Fraction of time residue is in contact with lipid.
DEFAULT_SURFACELIMIT = 0.05  # Surface (in square Angstroms) exposed to solvent.
DEFAULT_GRANULARITY  = 0.2   # Granularity (in Angstroms) of the histogram used to calculate the membrane planes.




# Global function to create an array initialised at a given length
#
def make_array(arrtype, length, initchar=0):
    a = array(arrtype)
    if type(a) == type(None):
        return
    i=0
    while i < length:
        a.append(initchar)
        i+=1
    return a


def is_entry_consistent(pdbrows, tablerows):
    if len(pdbrows) < len(tablerows):
        return False
    for i in xrange(len(tablerows)):
        t = tablerows[i]
        #p = pdbrows[t[Classifier.CGNUM]-1]
        p = pdbrows[i]
        #if p.res != t[Classifier.RES] or p.ires != t[Classifier.CGNUM] or (p.chain and (p.chain.upper() != t[Classifier.CHAIN].upper())):
        if p.res != t[Classifier.RES] or p.ires != t[Classifier.CGNUM] or (p.chain and (p.chain != t[Classifier.CHAIN])):
            return False
    return True


def get_entry_inconsistency(pdbrows, tablerows):
    if len(pdbrows) < len(tablerows):
        return "CGDB INCONSISTENCY: PDB and Table length differs: %d, %d\n" % (len(pdbrows), len(tablerows))
    output = ""
    for i in xrange(len(pdbrows)):
        t = tablerows[i]
        p = pdbrows[t[Classifier.CGNUM]-1]
        if   p.res != t[Classifier.RES]:
            output+="RESIDUE NAME: %s != %s\n"                   % (p.res, t[Classifier.RES])
        elif p.ires != t[Classifier.CGNUM]:
            output+="RESIDUE NUMBER & CG NUMBER: %s != %s\n"     % (p.ires, t[Classifier.CGNUM])
        #elif p.chain and (p.chain.upper() != t[Classifier.CHAIN].upper()):
        elif p.chain and (p.chain != t[Classifier.CHAIN]):
            output+="CHAIN: %s != NULL and %s != %s\n"           % (p.chain, p.chain, t[Classifier.CHAIN])
        else:
            continue
        output+=join(",  ", p)
        output+="\n"
        output+=join(",  ", t)
        output+="\n\n"
    return output


class NoDatabaseEntryError(NotFoundError):
    pass

class DatabaseInconsistencyError(ValueError):
    pass








re_imemid = re.compile(r"^(\d[a-zA-Z0-9]{3}\.\d+)(\..\.\d+)?$")

def get_main_entrycode(entry_code):
    m = re_imemid.match(entry_code)
    if not m:
      raise NoDatabaseEntryError("Entrycode not in a valid format: '%s'. Cannot retrieve file of type '%s'." % (entry_code, extension))

    code = m.group(1) # remove the chain identifier from the filename (there is only one table for all chains)
    return code  


def optimise_plane_limit(limit_min, limit_max, hist_class_below, hist_class_above):
    if limit_min > limit_max:
        raise ValueError("ERROR in optimise_plane_limit(): limit_min > limit_max !")
    elif limit_min == limit_max:
        return limit_min
    
    misclass = array('f')
    #corclass = array('f')
    # Try out various limits
    for limit in xrange(limit_min, limit_max):
        ## Count mis-classifications
        incorrect   = sum(hist_class_above[limit_min:limit]) + sum(hist_class_below[limit:limit_max])
        #correct = sum(hist_class_above[limit:limit_max]) + sum(hist_class_below[limit_min:limit])
        misclass.append(incorrect)
        #corclass.append(correct)
    
    value=misclass[0]
    index=0
    for i in xrange(1, len(misclass)):
        if misclass[i] < value:
            value = misclass[i]
            index = i

    # Thesis figures
    #print value,
    
    return index + limit_min


def parse_table(tablefile, pdb_data, offset):
    if not tablefile or not pdb_data:
        return

    # Easier debugging...
    assert 11 == Classifier.Z, "Column numbering inconsistent throughout python script!"

    table = []
    index = dict()
    for line in tablefile:
        line = line.strip()
        if not line or line[0]=='#':
            continue
        fields = line.split("\t")
        if len(fields) < 4:
            sys.stderr.write("Skipping line in CGDB table file for entry '%s':\n>>> %s\n" % (entry_code, line))
            continue

        # Turn numeric strings into numbers
        #
        fields[Classifier.CGNUM]  = int(fields[Classifier.CGNUM])
        fields[Classifier.RES]    = fields[Classifier.RES].upper()
        fields[Classifier.RESNUM] = int(fields[Classifier.RESNUM])
        #fields[Classifier.CHAIN]  = fields[Classifier.CHAIN].upper()
        fields[Classifier.CHAIN]  = fields[Classifier.CHAIN]

        #if fields[1:4] == prev[1:4]:
            #continue

        while len(fields) < 12:
            fields.append(0.0)

        # Read the Z coordinates from the PDB file, rather than the TABLE file
        #
        cgnum = fields[Classifier.CGNUM]
        
        strucindex = cgnum-offset-1
        if (strucindex < 0 or strucindex >= len(pdb_data)):
          continue

        i=4
        while i < 12:
            fields[i] = float(fields[i])
            i += 1

        pdbline = pdb_data[strucindex]
        fields[Classifier.Z] = pdbline.z
        index[cgnum] = len(table)
        fields = tuple(fields)
        table.append(fields)

    return table, index








    
class Classifier(object):
    annot_classes = ("membrane contact", "membrane layer", "accessibility", "membrane consensus", "membrane contact fraction")
    
    # Column indeces
    CGNUM, RES, RESNUM, CHAIN, FRACTION_ALL, FRACTION_HEAD, FRACTION_TAIL, MEAN_ALL, MEAN_HEAD, MEAN_TAIL, SASA, Z = range(12)
    
    
    def __init__(self, dbdir, entrycode, tablefile=None, pdbfile=None):
        if not entrycode:
            raise NoDatabaseEntryError("ERROR: Empty CGDB entry code given.")
        
        assert None != dbdir
        
        self.dbdir = dbdir
        self.entry_code = entrycode
        self.pdb_code   = entrycode[:4]
        
        self.entry_index = {}     # Maps CG numbers to list indeces in entry_data
        self.entry_data  = []     # Contains CGDB simulation results
        self.entry_data_all = None # Contains the full results table, even for residues not in pdb_data
        self.pdb_data    = []     # Content of the CGDB PDB file associated with this CGDB entry
        self.offset      = 0      # Difference between (residue number of the first residue) - 1
        self.class_strings = ()   # For mapping to the PDB file from CGDB (C-alpha and C-beta only)
        self.class_hashes  = ()   # For mapping to the PDB file from www.pdb.org
        self.planes        = None
        
        validate_pdb_code(self.pdb_code)
        
        if pdbfile:
          self.pdb_data = Pdb(self.entry_code, pdbfile).get_CA()
        else:
          try:
            self.pdb_data = Pdb(self.entry_code, self.get_file('pdb')).get_CA()
          except NoDatabaseEntryError:
            self.pdb_data = Pdb(self.entry_code, self.get_file('atm')).get_CA()
        
        self.offset = self.pdb_data[0].ires - 1
        
        #print "Offset:", self.offset
        
        if self.pdb_data[-1].ires != self.offset + len(self.pdb_data):
            for item in self.pdb_data:
                sys.stderr.write(str(item))
                sys.stderr.write("\n")
            sys.stderr.write("Length of pdb_data: %d\n" % (len(self.pdb_data)))
            raise DatabaseInconsistencyError, "ERROR: CGDB PDB numbering not continuous throughout for entry '%s'" % (self.entry_code)
        
        if tablefile == None:
            self.parse_table_file(self.get_file('table'))
            if not self.entry_data:
                raise NoDatabaseEntryError("ERROR: No CGDB entry '%s'" % (self.entry_code))
        else:
            self.parse_table_file(tablefile)
            if not self.entry_data:
                sys.stderr.write("WARNING: Could not successfully parse external PDB for entry '%s'\n" % (self.entry_code))
        
        self.crop_pdb_data()
        
        if not is_entry_consistent(self.pdb_data, self.entry_data):
            sys.stderr.write(get_entry_inconsistency(self.pdb_data, self.entry_data))
            raise DatabaseInconsistencyError("CGDB table inconsistent with CGDB PDB file: "+self.entry_code)
          
        
    
    # Make Classifier object seem like a list - content is the CGDB summary table
    #
    def __iter__(self):
        return self.entry_data.__iter__()
    def __getitem__(self, i):
        return self.entry_data.__getitem__(i)
    def __len__(self):
        return self.entry_data.__len__()
    def __str__(self):
        return "{Classifier:%s:%s}" % (self.entry_code, self.entry_data.__str__())
    
    
    
    def cg2index(self, cg):
        if not self.entry_index.has_key(cg):
            return -1
        return self.entry_index[cg]
    
    
    def get_chain_boundaries(self):
        "Returns the indeces of the rows where a new chain begins. Note that there may be duplicate chains - check the chain IDs in these rows!"
        if not len(self):
            return []
        
        i=0
        chainbreaks = []
        prevchain   = ''
        prevresnum  = 0
        for fields in self:
            c = fields[Classifier.CHAIN]
            n = fields[Classifier.RESNUM]
            if prevresnum > n or prevchain != c:
                chainbreaks.append(i)
                prevchain = c
            prevresnum = n
            i+=1

        return chainbreaks
    
    
    def crop_pdb_data(self):
        newdata = []
        for fields in self.pdb_data:
          cg = fields.ires
          idx = self.cg2index(cg)
          if idx>=0:
              newdata.append(fields)
        if len(newdata) < len(self.pdb_data):
          self.pdb_data = Pdb(self.pdb_data, newdata)
    
    
    def parse_table_file(self, tablefile):
        if not tablefile:
            return
        
        table, index = parse_table(tablefile, self.pdb_data, self.offset)
        maincode = get_main_entrycode(self.entry_code)
        if maincode != self.entry_code:
          tablefile.seek(0)
          fulltable, fullindex = parse_table(tablefile, Pdb(maincode, self.get_file('atm', entry_code=maincode)).get_CA(), 0)
        else:
          fulltable = table
        
        self.entry_data     = table
        self.entry_index    = index
        self.entry_data_all = fulltable
    
    
    def get_file(self, extension, entry_code=None):
        dbdir = self.dbdir
        if None==entry_code:
          entry_code = self.entry_code
        
        #print "get_file(%s, %s)" % (entry_code, extension)
        import os
        from glob import glob
        validate_pdb_code(entry_code[:4])
        
        insert = ""
        folder = "analysis"
        code = entry_code
        m = re_imemid.match(entry_code)
        if not m:
          raise NoDatabaseEntryError("Entrycode not in a valid format: '%s'. Cannot retrieve file of type '%s'." % (entry_code, extension))

        if extension in ("atm", "pdb"):
          if m.group(2): # Not a normal CGDB ID. Probably has a chain identifier, e.g. "1ABC0A"
            folder = "chains"
            #insert = ""
        elif extension == "table":
          code = m.group(1) # remove the chain identifier from the filename (there is only one table for all chains)
        
        filepattern = dbdir+'/'+folder+'/'+code+insert+'.'+extension
        matches = glob(filepattern)
        if not matches:
          raise NoDatabaseEntryError("Could not find file of type '%s' matching entry '%s': %s" % (extension, entry_code, filepattern))
        if len(matches) > 1:
          raise ValueError("More than one match for entry code '%s'" % (entry_code))
        return file(matches[0])

    
    
    def get_simple_planes(self, timeLimit):
        # Max and min z-coordinates for the three layer types
        #
        head_max = None
        head_min = None
        tail_max = None
        tail_min = None
        
        for fields in self.entry_data_all:
            contactAny   = fields[Classifier.FRACTION_ALL]
            contactHead  = fields[Classifier.FRACTION_HEAD]
            contactTail  = fields[Classifier.FRACTION_TAIL]
            zcoord       = fields[Classifier.Z]
            
            
            if   contactAny > timeLimit:
                if contactHead > timeLimit:
                    if   None == head_max or zcoord > head_max:
                        head_max = zcoord
                    elif None == head_min or zcoord < head_min:
                        head_min = zcoord
                if contactTail > timeLimit:
                    if   None == tail_max or zcoord > tail_max:
                        tail_max = zcoord
                    elif None == tail_min or zcoord < tail_min:
                        tail_min = zcoord
        
        
        if head_max == None:
            head_max = 0.0
        if head_min == None:
            head_min = 0.0
        if tail_max == None:
            tail_max = 0.0
        if tail_min == None:
            tail_min = 0.0
        
        
        if head_max < tail_max:
            head_max = tail_max
        if head_min > tail_min:
            head_min = tail_min
        
        return array('f', (head_max, tail_max, tail_min, head_min))
   
    
    def get_planes(self, timeLimit=DEFAULT_TIMELIMIT, surfaceLimit=DEFAULT_SURFACELIMIT, granularity=DEFAULT_GRANULARITY, weighted=True):
        if self.planes:
            return self.planes
        
        # granularity value is in angstrom
        
        limits = self.get_simple_planes(timeLimit) # , surfaceLimit
        
        #print "Simple Limits: ", limits
        
        # Max and min z-coordinates for the three layer types
        #
        hist_length = 2+int((limits[0]-limits[3])/granularity)
        hist_offset = limits[3]
        
        
        hist_head = make_array('f', hist_length)
        hist_tail = make_array('f', hist_length)
        hist_sasa = make_array('f', hist_length)
        
        max_sasa = 0
        for fields in self.entry_data_all:
            zcoord       = fields[Classifier.Z]
            
            if zcoord < limits[3] or zcoord > limits[0]:
                continue
            
            contactAny   = fields[Classifier.FRACTION_ALL]
            contactHead  = fields[Classifier.FRACTION_HEAD]
            contactTail  = fields[Classifier.FRACTION_TAIL]
            sasa         = fields[Classifier.SASA]
            
            index = int((zcoord-hist_offset)/granularity)
            
            if weighted:
                if contactAny > timeLimit:
                    hist_head[index] += contactHead
                    hist_tail[index] += contactTail
                elif sasa > surfaceLimit:
                    hist_sasa[index] += sasa
                    max_sasa = max(max_sasa, sasa)
            else:
                if contactAny > timeLimit:
                    if contactHead > contactTail:
                        hist_head[index] += 1
                    else:
                        hist_tail[index] += 1
                elif sasa > surfaceLimit:
                        hist_sasa[index] += 1
        
        if weighted and max_sasa > 0:
            for i in range(len(hist_sasa)):
                hist_sasa[i] /= max_sasa
        
        limit_indeces = array('i')
        for zcoord in limits:
            limit_indeces.append(int((zcoord-hist_offset)/granularity))
        
        #print "Limit indices:", limit_indeces
        
        #hist_nottail = [hist_head[i]+hist_sasa[i] for i in xrange(hist_length)]
        #hist_nothead = [hist_tail[i]+hist_sasa[i] for i in xrange(hist_length)]
        hist_notsasa = [hist_head[i]+hist_tail[i] for i in xrange(hist_length)]
        
        #print self.entry_code,
        
        # Find boundary between membrane and non-membrane
        limit_indeces[0] = optimise_plane_limit(limit_indeces[3], limit_indeces[0], hist_notsasa, hist_sasa) # move down top outer membrane limit
        limit_indeces[3] = optimise_plane_limit(limit_indeces[3], limit_indeces[0], hist_sasa, hist_notsasa) # move up bottom outer membrane limit
        
        # Find boundary (within membrane region) between inner and outer membrane layers
        limit_indeces[1] = optimise_plane_limit(limit_indeces[3], limit_indeces[0], hist_tail, hist_head) # move down top inner membrane limit
        limit_indeces[2] = optimise_plane_limit(limit_indeces[3], limit_indeces[0], hist_head, hist_tail) # move up bottom inner membrane limit
        
        #print
        
        for i in xrange(len(limits)):
            limits[i] = hist_offset+float(limit_indeces[i])*granularity
        
        # Thesis figures
        #print self.entry_code
        #for i in xrange(len(hist_head)):
        #  print i, hist_sasa[i], hist_head[i], hist_tail[i]
        #print
        
        #print "Adjusted Limits: ", limits
        
        self.planes = limits
        return limits
    
    
    def get_class_strings(self):
        assert self.class_strings, "Call calculate_classes() first"
        return self.class_strings
    
    
    def z2class(self, zcoord):
      assert self.planes, "Call calculate_classes() or get_planes() before calling z2class()!"
      
      if   zcoord > self.planes[0] or zcoord < self.planes[3]:
          # Residue is OUTSIDE the membrane, in an aqueous environment
          return 'N'
      elif zcoord > self.planes[1] or zcoord < self.planes[2]:
          # Residue is in the HEAD group layer of the membrane
          return 'H'
      else:
          # Residue is in the TAIL group layer of the membrane
          return 'T'
    
    
    def calculate_classes(self, timeLimit=DEFAULT_TIMELIMIT, surfaceLimit=DEFAULT_SURFACELIMIT, granularity=DEFAULT_GRANULARITY):
        # Clear data from previous calculations
        self.class_hashes = ()
        self.planes = None
        self.get_planes(timeLimit, surfaceLimit, granularity)
        
        residue_contact   = array('c')
        residue_layer     = array('c')
        residue_access    = array('c')
        residue_consensus = array('c')
        residue_contfrac  = array('c')
        
        i=-1
        for fields in self:
            i+=1
            #res          = fields[2]
            #chain        = fields[3]
            
            contactAny   = fields[Classifier.FRACTION_ALL]
            contactHead  = fields[Classifier.FRACTION_HEAD]
            contactTail  = fields[Classifier.FRACTION_TAIL]
            sasa         = fields[Classifier.SASA]
            zcoord       = fields[Classifier.Z]
            
            if sasa > surfaceLimit:
              residue_access.append('T')
            else:
              residue_access.append('F')
            
            if contactAny > timeLimit:
#            if contactHead > timeLimit or contactTail > timeLimit:
                if contactHead > contactTail:
                    residue_contact.append('H')
                    x = min(max(0, int(contactHead*10)), 9)
                    residue_contfrac.append(str(x))
                else:
                    residue_contact.append('T')
                    x = min(max(0, int(contactTail*10)), 9)
                    residue_contfrac.append(str(x))
            else:
                residue_contact.append('N')
                residue_contfrac.append('0')
            
            residue_layer.append(self.z2class(zcoord))
            
            # Make consensus of contact, layer and accessibility:
            #
            x = residue_layer[-1]
            if residue_access[-1] == 'T':
              if residue_contact[-1] == 'N' and residue_layer[-1] == 'H':
                x = 'y'
              elif residue_contact[-1] == 'N' and residue_layer[-1] == 'T':
                x = 'p'
              elif residue_contact[-1] == 'H' or residue_layer[-1] == 'H':
                x = 'h'
              else:
                x = residue_contact[-1].lower()
            residue_consensus.append(x)
        
        self.class_strings = (residue_contact.tostring(), residue_layer.tostring(), residue_access.tostring(), residue_consensus.tostring(), residue_contfrac.tostring())
    

class Builder(object):
    
    def __init__(self, alif, classifier=None, dbdir=None):
        assert alif, "Need to provide input data to Builder()!"
        
        self.ali        = alif
        self.classifier = classifier
        self.annot_classes = []

        if not self.classifier:
          for egroup in self.ali:
            try:
              e = egroup["structure"]
              self.classifier = Classifier(dbdir, e.getCode())
              break
            except IndexError:
              pass
            except NoDatabaseEntryError:
              pass
        if not self.classifier:
          raise NoDatabaseEntryError("No Classifier found for Builder")
    
    
    def write_output(self, outdir, subdir=None):
        entrycode = self.classifier.entry_code
        if not subdir:
          subdir = entrycode
        
        entrydir = "%s/%s"%(outdir, subdir)
        if not os.path.exists(entrydir):
            os.mkdir(entrydir)
        
        struc = self.classifier.pdb_data.get_root_object()
        planes = self.classifier.get_planes()
        
        
        f = open(os.path.join(entrydir, "annotation.tem"), 'w')
        f.write(str(self.ali))
        f.close()
        
        f = open(os.path.join(entrydir, "sequence.fasta"), 'w')
        f.write(self.ali[entrycode].master.toFastaString(remove_gaps=True))
        f.close()
        
        f = open(os.path.join(entrydir, "layer.txt"), 'w')
        f.write(join(" ", planes))
        f.write("\n")
        f.close()
        
        for entry in self.ali[entrycode]:
          if not entry.isMaster():
            assert entry.desc in Classifier.annot_classes
            
            if entry.desc not in ("membrane contact", "membrane layer", "membrane consensus"):
              continue
            
            cl = entry.getType()
            class_string = deGappify(entry.get_seq())
            self.annotate_structure(struc, class_string)
            
            shortclass = cl.split()[-1]
            
            f=open(os.path.join(entrydir, shortclass+".atm"), 'w')
            f.write(str(struc))
            f.close()
            
            if cl=="membrane contact":
              os.system('ln -sf "%s.atm" "%s/%s.atm"'%(shortclass, entrydir, entrycode))
        
        os.system('chmod a+r "%s/"*.*' % (entrydir))
        os.system('chmod a+rx "%s"' % (entrydir))
        
    
    
    # Instance methods
    #
    
    def make_all_basic_annotation(self, timeLimit=DEFAULT_TIMELIMIT, surfaceLimit=DEFAULT_SURFACELIMIT, granularity=DEFAULT_GRANULARITY, replace=False):
        clsfr = self.classifier
        assert clsfr
        
        clsfr.calculate_classes(timeLimit, surfaceLimit, granularity)
        class_strings = list(clsfr.get_class_strings())

        templateseq = self.ali[clsfr.entry_code].getMasterEntry().get_seq()
        for i,s in enumerate(class_strings):
          try:
              class_strings[i] = gappify(templateseq, s)
          except ValueError:
              sys.stderr.write("ERROR gappifying class string for entry:\n%s\n%s\n\nTarget annotation:\n%s\n" % (self.classifier.entry_code, templateseq, s));
              raise
        
        output = [] # empty annotation list
        self.annot_classes = [] # empty annotation class list
        
        for entrygroup in self.ali:
            masterentry = entrygroup.getMasterEntry()
            
            for i,s in enumerate(class_strings):
                clname = Classifier.annot_classes[i]
                myannot=array('c')
                for j,c in enumerate(s):
                  if not isGap(masterentry.seq[j]):
                    if isGap(c):
                      myannot.append('?')
                    else:
                      myannot.append(c)
                  else:
                    myannot.append('-')
                output.append(Ali.Entry(masterentry.code, clname, myannot.tostring()))
        
        self.annot_classes.extend(Classifier.annot_classes[:len(class_strings)])
        
        for entry in output:
          if not self.ali.add_entry(entry, ignore_duplicates=True):
            if replace:
              self.ali[entry.getCode()][entry.getType()] = entry
    
    annotate = make_all_basic_annotation
    
    
    def annotate_structure(self, struc, class_string):
        assert len(class_string) == struc.rescount()
        
        for i,res in enumerate(struc.xresidues()):
          c = class_string[i]
          b = 0.0
          if c in "Nnyp":
              b = 1.0
          elif c in "Hh":
              b = 100.0
          elif c in "Tt":
              b = 200.0
          else:
              b = 0.0
          for atm in res:
            atm.b = b
