#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# iMembrane : search executable
#
# Author: Sebastian Kelm
# Created: 10/05/2010 - rewritten from original Perl executable
#

import sys
import os
import re
import tempfile
import subprocess
from glob import glob
from operator import itemgetter
#from threading import Thread

from prosci.common import IllegalStateError
from prosci.util.ali import Ali
from prosci.imembrane.projector import Projector, imemid2pdb, PoorSuperpositionError


re_nonword = re.compile("[^a-zA-Z0-9_\-,.]")
re_error = re.compile("error", re.IGNORECASE)

def inputid_hash(inputid):
  if re_nonword.search(inputid):
    return re_nonword.sub("_", inputid)
  return inputid


class SearchHit:
    def __init__(self, p):
      self.scores = p.scores
      self.dbid = p.dbid
      self.inputid = p.inputid
      
      if p.struc:
        self.__cmp__ = self.__cmp_struc
      elif not "E-value" in self.scores:
        self.__cmp__ = self.__cmp_seq
      else:
        self.__cmp__ = self.__cmp_evalue
      
      self.subdir = "%s_%s"%(inputid_hash(p.inputid), p.dbid)

    def __cmp_struc(self, other):
      if other is None:
        return 1
      return cmp(other.scores["TM-score"]*other.scores["pannot"], self.scores["TM-score"]*self.scores["pannot"])
      
    def __cmp_seq(self, other):
      if other is None:
        return 1
      return cmp(other.scores["pid"]*other.scores["pannot"], self.scores["pid"]*self.scores["pannot"])
    
    def __cmp_evalue(self, other):
      if other is None:
        return 1
      return cmp(self.scores["E-value"], other.scores["E-value"])


class Searcher:
    def __init__(self, dbdir, inputid, seq, struc, seqprofile=None, outdir=None, maxhits=None, hitlog="hits.table", verbose=False, variants=False, exhaustive=False, fast=False, write_input_sequence=False, run_joy=False, transmembrane=False, outputparent = "copy"):
      
      if None==outdir:
        outdir = os.getcwd()
      elif not os.path.exists(outdir):
        os.mkdir(outdir)
      
      if hitlog == '-':
        hitlog = None;
      
      if exhaustive and fast:
        raise ArgumentError("Must not provide both 'exhaustive' and 'fast' options")
      
      if seq and isinstance(seq, basestring):
        seq = Ali(file(seq), ignore_duplicates=True, fasta_mode=True)
      
      if outputparent:
        outputparent = outputparent.lower()
        assert outputparent in ("copy", "link", "none")
      else:
        outputparent = None
      self.outputparent = outputparent
      
      self.dbdir = dbdir
      self.inputid = inputid
      self.seq = seq
      self.struc = struc
      self.seqprofile = seqprofile
      
      self.outdir = outdir
      self.hitlog = hitlog
      self.maxhits = maxhits
      self.verbose = verbose
      self.variants = variants
      self.exhaustive = exhaustive
      self.fast = fast
      self.write_input_sequence = write_input_sequence
      self.run_joy = run_joy
      self.transmembrane = transmembrane
      
      self.dummy = Projector(dbdir, None, inputid, seq, struc)
      
      self.hits = []
    
    
    def get_representative_set(self, minscore=0.0):
      try:
        f = open(os.path.join(self.dbdir, "representatives.table"))
      except:
        return set([])
      
      result = []
      try:
        for line in f:
          pdb, pdbsim, tmscore = line.split()
          if float(tmscore) >= minscore:
            result.append(pdbsim)
        return set(result)
      finally:
        f.close()
    
    
    def search(self, iterations = 3, cutoff_e = 1.0, cutoff_e2 = 0.002, cutoff_tm = 0.5, cutoff_rmsd = 10.0, cutoff_pid = 0.00, cutoff_pcov = 0.50, cutoff_pannot = 0.50, cutoff_representative = 0.50):
      
      verbose = self.verbose
      dummy = self.dummy
      seqprofile = self.seqprofile
      
      
      def fast_search(evalue=None, hvalue=None, iterations=None):
          dbid = None
          p    = Projector(self.dbdir, dbid, self.inputid, self.seq, self.struc)
          
          inputid  = p.inputid
          inputseq = p.ali[inputid].getMasterEntry().get_seq()
          
          if seqprofile:
            f = file(seqprofile)
            profile_ali = f.read()
            f.close()
            profile_ali = Ali(profile_ali, fasta_mode=True)
            aligned = True
            for eg in profile_ali:
              if len(eg.master.seq) != len(profile_ali[0].master.seq):
                aligned = False
                break
            if not aligned:
              profile_ali.align()
          
          
          if not iterations:
            iterations = 3
          
          if not evalue:
            evalue = 1.0
          else:
            evalue = float(evalue)
          
          if not hvalue:
            hvalue = min(0.002, evalue/10)
          else:
            hvalue = float(hvalue)
          
          dbpath = "%s/seq/chains"%(self.dbdir)
          assert glob("%s.p*"%dbpath)
          
          
          fd, tmpfname = tempfile.mkstemp()
          f = os.fdopen(fd, 'w')
          f.write(">%s\n%s\n"%(inputid, inputseq))
          if seqprofile:
            f.write(str(profile_ali))
          f.close()
          
          
          try:
            if not seqprofile:
              p = subprocess.Popen("psiblast -query '%s' -db '%s' -num_iterations %d -evalue %f -inclusion_ethresh %f -outfmt 6" % (tmpfname, dbpath, iterations, evalue, hvalue), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
              result, errors = p.communicate()
            else:
              p = subprocess.Popen("psiblast -in_msa '%s' -db '%s' -num_iterations %d -evalue %f -inclusion_ethresh %f -outfmt 6" % (tmpfname, dbpath, iterations, evalue, hvalue), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
              result, errors = p.communicate()
          finally:
            os.remove(tmpfname)
          
          for l in errors.splitlines():
            if re_error.search(l):
              raise IllegalStateError("BLAST crashed: '%s'" % (l))
          
          resultlist = []
          for l in result.splitlines():
            fields = l.strip().split()
            try:
              fields = [fields[0], fields[1], float(fields[10])]
            except IndexError:
              continue
            except ValueError:
              continue
            resultlist.append(fields)
          
          resultlist.sort(key=itemgetter(2), reverse=True)
          
          resultdict = {}
          for fields in resultlist:
            match_id     = fields[1]
            match_evalue = fields[2]
            resultdict[match_id] = match_evalue
          
          return resultdict
        
        
        
      def main_search(searchids, evalues):
          
          hitproj = [] # list of Projector objects corresponding to each hit
          hitinfo = [] # list of ids and scores corresponding to each hit
          
          def sort_and_prune_hits():
            # If we have structure information, sort by structure similarity (* % annotated)
            # otherwise, if we have done a BLAST search, leave it sorted by BLAST E-value
            # otherwise sort by sequence identity (* % annotated)
            if self.struc:
              hitproj.sort(key=lambda x: -x.scores["TM-score"]*x.scores["pannot"])
            elif not evalues:
              hitproj.sort(key=lambda x: -x.scores["pid"]*x.scores["pannot"])
            else:
              hitproj.sort(key=lambda x: x.scores["E-value"])
            
            if self.maxhits:
              del hitproj[self.maxhits:]
              assert len(hitproj) <= self.maxhits
          
          def write_hit(p):
            hit = SearchHit(p)
            p.write_output(self.outdir, hit.subdir, copy_parent=(self.outputparent=="copy"), link_parent=(self.outputparent=="link"), write_fasta=self.write_input_sequence)
            hitinfo.append(hit)
          
          
          # Speed optimisation. Keeps track of rejected pdb chains.
          # This speeds up the search if "variants" are allowed.
          rejected = set([])
          
          
          # Do the actual search
          #
          for i, dbid in enumerate(searchids):
            
            if self.variants:
              repid = imemid2pdb(dbid)
              if repid in rejected:
                if verbose:
                  print "Skipping  hit %6d / %6d : %s" % (i+1, len(searchids), dbid)
                continue
            
            try:
              p = Projector(self.dbdir, dbid, dummy.inputid, dummy.ali, dummy.struc, run_joy=self.run_joy)
            except PoorSuperpositionError:
              if self.variants:
                rejected.add(repid)
              continue
            
            if evalues:
              p.scores["E-value"] = evalues[dbid]
            else:
              p.scores["E-value"] = -1.0
            
            if (dummy.struc and p.scores["TM-score"] < cutoff_tm)   or \
               (dummy.struc and p.scores["rmsd"]     > cutoff_rmsd) or \
               (                p.scores["pid"]      < cutoff_pid)  or \
               (                p.scores["pcov"]     < cutoff_pcov):
              if verbose:
                #print "Thread %s : Rejecting hit %6d / %6d : %s" % (self.threadnum, i+1, self.total_length, dbid)
                print "Rejecting hit %6d / %6d : %s" % (i+1, len(searchids), dbid)
              if self.variants:
                rejected.add(repid)
              continue
            
            p.annotate()
            
            if ("pannot"   in p.scores and p.scores["pannot"]   < cutoff_pannot):
              if verbose:
                print "Rejecting hit %6d / %6d : %s" % (i+1, len(searchids), dbid)
              if self.variants:
                rejected.add(repid)
              continue
            
            if self.transmembrane and ("T" not in p.ali[p.inputid]["membrane layer"].seq):
              if verbose:
                print "Rejecting hit %6d / %6d : %s : Reaches specified cut-offs but is not a transmembrane domain" % (i+1, len(searchids), dbid)
              continue
            
            if verbose:
              print "Saving    hit %6d / %6d : %s" % (i+1, len(searchids), dbid)
            
            if not self.maxhits:
              write_hit(p)
            else:
              hitproj.append(p)
              if len(hitproj) >= self.maxhits + 10:
                # Keep memory usage low by periodically throwing out the 10 worst hits
                sort_and_prune_hits()
          
          if self.maxhits:
            sort_and_prune_hits()
            for p in hitproj:
              write_hit(p)
          else:
            hitinfo.sort()
          
          del hitproj[:]
          return hitinfo
        
      
      
      
      ## Do a fast search, to establish possible candidates
      
      
      searchids = []
      evalues   = None
      if not self.exhaustive:
        try:
          evalues = fast_search(cutoff_e, cutoff_e2, iterations)
        except IllegalStateError, e:
          sys.stderr.write(str(e))
          sys.stderr.write("\nCarrying on as if BLAST search had yielded no hits...\n")
        
        if evalues:
          searchids = evalues.keys()
          searchids.sort(key=lambda x: evalues[x])
          if verbose:
            print "BLAST search: successful"
        elif verbose:
            print "BLAST search: no hits"
      
      
      ## If no hits found with a fast search, consider doing an exhaustive search
      
      def prep_exhaustive_search(searchids):
        if verbose:
          print "Performing exhaustive search"
        for entrydir in glob("%s/entries/*"%self.dbdir):
          searchids.append(os.path.basename(entrydir))
      
      
      def filter_representatives(searchids):
        # Do we want to allow all the variants (multiple simulations per
        # protein chain) as hits?
        #
        if not self.variants:
          if verbose:
            print "Allowing only one db entry per PDB chain"
          
          searchids_new = []
          representatives = self.get_representative_set(cutoff_representative)

          if representatives:
            # We have a list of representatives
            # Intersect the searchid list with the list of representatives
            for e in searchids:
              efields = e.split('.')
              pdbsim = efields[0]+'.'+efields[1]
              if pdbsim in representatives:
                searchids_new.append(e)
          else:
            # We have no list of representatives
            # Just use the first simulation for each protein chain
            representatives = set([])
            for e in searchids:
              repid = imemid2pdb(e)
              if repid not in representatives:
                representatives.add(repid)
                searchids_new.append(e)
          
          del searchids[:]
          searchids.extend(searchids_new)
          del representatives
          
        else:
          if verbose:
            print "Allowing multiple db entries per PDB chain"

        if verbose:
          print
        
        
      
      
      if not searchids and not self.fast:
        prep_exhaustive_search(searchids)
      
      filter_representatives(searchids)
      
      ## We have now established which database entries we want to try and project onto our query
      
      self.hits = main_search(searchids, evalues)
      
      
      if evalues and (not self.hits) and (not self.fast):
        # We're auto-detecting the search type
        # We've done a fast search and found some stuff but none of the candidates were suitable
        # Let's reset everything and just do an exhaustive search...
        #
        searchids = []
        evalues   = None
        prep_exhaustive_search(searchids)
        filter_representatives(searchids)
        self.hits = main_search(searchids, evalues)
      
      
      ## We've done the search. Files have already been written. Now output the list of hits.
      
      
      if self.hits:
        score_columns = sorted(self.hits[0].scores.keys())
        
        if self.hitlog:
          if self.hitlog == os.path.basename(self.hitlog):
            # If just a bare filename (without a path) has been provided, put
            # the logfile in the output directory.
            hitlog_stream = open(os.path.join(self.outdir, self.hitlog), 'w')
          else:
            hitlog_stream = open(self.hitlog, 'w')
        else:
          hitlog_stream = sys.stdout
          if verbose:
            hitlog_stream.write("\nHits:\n")
        
        hitlog_stream.write("#subdirectory\tinputid\tdbid\t%s\n" % ("\t".join(score_columns)))
        
        for hit in self.hits:
          hitlog_stream.write("%s\t%s\t%s" % (hit.subdir, hit.inputid, hit.dbid))
          for col in score_columns:
            hitlog_stream.write("\t%g" % (hit.scores[col]))
          hitlog_stream.write("\n")
        
        if self.hitlog:
          hitlog_stream.close()
        
        # Make a symbolic link to the top-scoring hit, for convenience.
        os.system("ln -sfn %s %s/tophit"%(self.hits[0].subdir, self.outdir))

      return self.hits
