#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import tempfile
import shutil
import subprocess
import shlex
import copy
from glob import glob

from prosci.common import write_file, IllegalStateError, which
from prosci.util.ali import Ali
from prosci.util.protein import Protein, ResidueList
from prosci.util.gaps import deGappify
from prosci.util.pdb3d import dist

MODELLER_BINARY_INSTALLED = bool(which("modeler"))
MODELLER_LIBARY_INSTALLED = False
try:
  from modeller import *
  from modeller.automodel import *
  MODELLER_LIBARY_INSTALLED = True
except ImportError:
  pass


if not MODELLER_BINARY_INSTALLED and not MODELLER_LIBARY_INSTALLED:
  raise ImportError("Modeller not installed")


SCRIPT_TEMPLATE_START = """
from modeller import *
from modeller.automodel import *

log.level(output=0, notes=0, warnings=0, errors=1, memory=0)
env = environ()
env.io.atom_files_directory = %s
"""

SCRIPT_TEMPLATE_QUICK = """
a = automodel(env,
                  alnfile = %s,
                  sequence = %s,
                  knowns = %s)
a.starting_model = 1
a.ending_model = 1
a.very_fast()
a.make()
"""

SCRIPT_TEMPLATE_REFINE = """
refine_sidechains = %s
refine_selection = %s
class LocallyRefinedModel(automodel):
    def select_atoms(self):
      if not self.refine_selection and not self.refine_sidechains:
        # No gaps and no sidechains, i.e. nothing to refine.
        # However, selection may not be empty.
        # We will use .very_fast() mode, so what we select doesn't matter.
        # Just select everything.
        return selection(self)
      
      # Select any gap residues and (if requested) all sidechains for optimisation.
      #
      sele = selection()
      for res in self.refine_selection:
        sele.add(self.residues[res])
      if self.refine_sidechains:
        return sele | selection(self).only_sidechain()
      else:
        return sele

a = LocallyRefinedModel(env,
    alnfile = %s,
    sequence = %s,
    knowns = %s,
    inifile = %s)
a.refine_selection = refine_selection
a.refine_sidechains = refine_sidechains

if not refine_selection and not refine_sidechains:
  # We don't have gaps and aren't refining sidechains.
  # Just do an unrefined model (i.e. do pretty much nothing).
  a.very_fast()

a.starting_model = 1
a.ending_model = 1
a.make()
"""


def filter_model_files(filenames):
    output = []

    has_fit_suffix = False
    for fname in filenames:
      if fname.endswith("_fit.pdb"):
        has_fit_suffix = True
        break
    
    for fname in filenames:
      if not has_fit_suffix or fname.endswith("_fit.pdb"):
        if "B9999" in os.path.basename(fname) and fname.endswith(".pdb"):
          output.append(fname)
    
    return output


def runModeller(scripttemplate, scriptvariables, template_directories=None):
    "Run a python script via the executable version of MODELLER"
    
    scripttext = scripttemplate % tuple([repr(x) for x in scriptvariables])
    
    if template_directories is not None:
      if isinstance(template_directories, basestring):
        template_directories = template_directories.strip(':').split(':')
      for i, d in enumerate(template_directories):
         template_directories[i] = os.path.abspath(d)
    
    prevdir = os.getcwd()
    searchdir = ['.', prevdir]
    if template_directories:
      searchdir += template_directories
    
    scripttext = (SCRIPT_TEMPLATE_START % (repr(searchdir))) + scripttext
    
    outdir = tempfile.mkdtemp()
    try:
      if not MODELLER_LIBARY_INSTALLED:
        scriptfile = os.path.join(outdir, "modelme.py")
        write_file(scriptfile, scripttext)
        
        p = subprocess.Popen(shlex.split("modeler %s"%(scriptfile)), cwd=outdir, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
      else:
        oldstdout = sys.stdout
        sys.stdout = open(os.devnull, "wb")
        os.chdir(outdir)
        try:
          exec scripttext
        finally:
          os.chdir(prevdir)
          sys.stdout.close()
          sys.stdout = oldstdout
      output = [ResidueList(x) for x in sorted(filter_model_files(glob(os.path.join(outdir, "*.pdb"))))]
      if not output:
        raise IllegalStateError("Modeller failed to generate a model:\n"+err)
    finally:
      shutil.rmtree(outdir)
    
    return output


def createModel(alignmentFileName, targetid, template_ids, template_directory=None, outFileName=None):
  alignmentFileName = os.path.abspath(alignmentFileName)
  
  if outFileName!=None:
    outFileName = os.path.abspath(outFileName)
  
  models = runModeller(SCRIPT_TEMPLATE_QUICK, (alignmentFileName, targetid, template_ids), template_directory)
  
  outputModel = Protein(models[0])
  outputModel.renumber()
  outputModel = outputModel.to_residuelist()
  outputModel.code = targetid
  
  if outFileName != None:
    write_file(fname, str(outputModel))
  
  return outputModel


def refineModel(inputModelFileName, refine_residues, alignmentFileName, targetid, template_ids, template_directory=None, outFileName=None, refine_sidechains=False):
  inputModelFileName = os.path.abspath(inputModelFileName)
  alignmentFileName  = os.path.abspath(alignmentFileName)
  
  models = runModeller(SCRIPT_TEMPLATE_REFINE, (refine_sidechains, refine_residues, alignmentFileName, targetid, template_ids, inputModelFileName), template_directory)
  
  outputModel = Protein(models[0])
  outputModel.renumber()
  outputModel = outputModel.to_residuelist()
  outputModel.code = targetid
  
  if outFileName != None:
    write_file(fname, str(outputModel))
  
  return outputModel


def writeAlignmentInModellerFormat(fname, alignment, targetid):
  headerSeq   = ">P1;%s\nsequence:::::::::\n"
  headerStruc = ">P1;%s\nstructure:%s:FIRST:@:END::unknown:unknown:-1:-1\n"
  
  f = open(fname, 'w')
  try:
    for eg in alignment:
      e = eg.master
      
      #if code == targetid:
        #f.write(headerSeq%(code))
      #else:
      f.write(headerStruc%(e.code, e.code))
      f.write(e.seq.upper())
      f.write("*\n")
  finally:
    f.close()


def mergeModels(mainModel, partialModel, target_residue_indices, bfactor=None):
  if bfactor is not None:
    for r in mainModel:
      for atm in r:
        atm.b = bfactor
  
  assert len(partialModel) <= len(mainModel)
  
  """
  # Some sanity checking...
  #
  gapsum=gaplist[0][1]
  seqsum=0
  for ix, g in enumerate(gaplist[1:]):
    seqsum += g[0] - gaplist[ix][0]
    gapsum += g[1]
  
  assert seqsum == len(partialModel), "%d (seq) != %d (partial model)" % (seqsum, len(partialModel))
  assert seqsum + gapsum == len(mainModel), "%d (seq) + %d (gap) != %d (total)" % (seqsum, gapsum, len(mainModel))
  """
  
  # Now let's merge those models!
  #
  """
  gapsum=gaplist[0][1]
  fragmentstart=0
  for ix, g in enumerate(gaplist[1:]):
    fragmentstart = gaplist[ix][0]
    gapstart = g[0]
    gaplen   = g[1]
    
    for i in xrange(fragmentstart, gapstart):
      try:
        assert mainModel[i+gapsum].ires == partialModel[i].ires, str(mainModel[i+gapsum]) + "\n" + str(partialModel[i])
      except IndexError:
        print str(mainModel[i+gapsum:i+gapsum+1])
        print str(partialModel[i:i+1])
        raise
      chaincode = mainModel[i+gapsum].chain
      mainModel[i+gapsum] = partialModel[i].copy()
      mainModel[i+gapsum].chain = chaincode
    
    gapsum += gaplen
  
  assert len(mainModel) == len(partialModel) + gapsum
  """
  
  for r, i in zip(partialModel, target_residue_indices):
    try:
      assert mainModel[i].res == r.res, str(mainModel[i]) + "\n" + str(r)
    except IndexError:
      print str(mainModel[i:i+1])
      print str(r)
      raise
    chaincode = mainModel[i].chain
    mainModel[i] = r.copy()
    mainModel[i].chain = chaincode
  
  newModel = copy.deepcopy(mainModel)
  newModel.renumber_atoms()
  
  #for r in newModel:
  #  for atm in r:
  #    atm.chain = "A"
  
  return newModel


"""
def extractLoopModel(mainModel, gaplist):
  "Extract only loop regions specified by the gaplist and put them into a new PDB file"
  loopModel       = ResidueList([], code="loop_model")
  
  gapsum = 0
  for g in gaplist:
    gapstart = g[0]
    gaplen   = g[1]
    for i in xrange(gapsum+gapstart, gapsum+gapstart+gaplen):
      loopModel.append(mainModel[i])
    gapsum += gaplen
  return loopModel
"""


def completeModel(myModel, alignment, targetid=None, template_ids=None, template_directory=None, outFileName=None, bfactor=20, refine_sidechains=False, chaincode=None, debug=False):
  "Complete a model using MODELLER"
  
  if not isinstance(myModel, Protein):
    myModel = Protein(myModel)
  # Save a protein chain codes
  myModel_chaincodes = [x.chain for x in myModel]
  
  myModel = myModel.to_residuelist()
  
  completeSeq = deGappify(alignment[targetid].master.seq)
  target_residue_indices = myModel.map_to_seq(completeSeq)
  
  
  if targetid==None:
    targetid=myModel.code
  
  if template_directory is not None:
    if isinstance(template_directory, basestring):
      template_directory = template_directory.strip(':').split(':')
    for i, dir in enumerate(template_directory):
       template_directory[i] = os.path.abspath(dir)
  
  if not template_ids:
    template_ids=[]
    for eg in alignment:
      code = eg.getCode()
      if code != targetid:
        template_ids.append(code)
  
  if outFileName!=None:
    outFileName = os.path.abspath(outFileName)
  
  prevdir = os.getcwd()
  ## DEBUG
  #
  if debug:
    from datetime import datetime
    now = str(int((datetime.now()-datetime(1970,1,1)).total_seconds()))[-5:]
    outdir = tempfile.mkdtemp(prefix=now)
    print "Saving intermediate MODELLER output to: %s"%(outdir)
  else:
    outdir = tempfile.mkdtemp()
  
  try:
    
    os.chdir(outdir)
    
    if not template_ids:
        alignment = copy.deepcopy(alignment)
        newentry = Ali.Entry(alignment[targetid].getMasterEntry())
        newentry.code = "template"
        newentry.desc = "structure"
        alignment.add_entry(newentry)
        
        template_ids.append(newentry.code)
        
        write_file("template.pdb", str(myModel))
    
    writeAlignmentInModellerFormat("model_alignment.ali", alignment, targetid)
    
    if debug:
      write_file("myModel.pdb", str(myModel))

    if len(myModel) < len(completeSeq):
      modellerModel  = createModel("model_alignment.ali", targetid, template_ids, template_directory)
      write_file("modellerModel.pdb", str(modellerModel))
      assert len(modellerModel) == len(completeSeq), "\n%s\n%s\n" % (modellerModel.get_seq(), completeSeq)
      compositeModel = mergeModels(modellerModel, myModel, target_residue_indices, bfactor)
    else:
      compositeModel = myModel
    
    write_file("compositeModel.pdb", str(compositeModel))
    
    
    refine_residues = set(range(len(completeSeq))).difference(target_residue_indices)
    refine_residues = sorted(refine_residues)
    
    finalModel = refineModel("compositeModel.pdb", refine_residues, "model_alignment.ali", targetid, template_ids, template_directory, refine_sidechains=refine_sidechains)
    
    if debug:
      write_file("finalModel.pdb", str(finalModel))
    
    
    assert len(compositeModel) == len(finalModel)
    for i in xrange(len(compositeModel)):
      r_comp = compositeModel[i]
      r_final = finalModel[i]
      if r_comp.CA.b == bfactor:
        # Refined portions: Add bfactor as an offset to all Modeller B factors
        for a in r_final:
          a.b += bfactor
      else:
        # Unrefined portions: Keep all B factors from myModel
        for a in r_final:
          a.b = r_comp.CA.b
    
    
    # Superimpose output model onto input model, excluding residues that were refined
    #
    if refine_residues:
      from prosci.util.pdb3d import superimpose
      subset_in = ResidueList([compositeModel[i] for i in target_residue_indices])
      subset_out = ResidueList([finalModel[i] for i in target_residue_indices])
      superimpose(finalModel.to_pdb(), compositeModel.to_pdb(), subset1=subset_out.to_pdb(), subset2=subset_in.to_pdb())
    
    finalModel = Protein(finalModel)
    
    
    # Assign chain codes
    if chaincode is not None:
      for chain in finalModel:
        chain.chain = chaincode
    elif len(myModel_chaincodes) == len(finalModel):
      for mychaincode, chain in zip(myModel_chaincodes, finalModel):
        chain.chain = mychaincode
    
    if outFileName != None:
      write_file(outFileName, str(finalModel))
    
    #loopModel = extractLoopModel(finalModel, gaplist)
    
    
  finally:
    os.chdir(prevdir)
    if not debug:
      shutil.rmtree(outdir)
  
  return Protein(finalModel) #, loopModel

